#include "Config_YM7101.h"
#include "DisplayAdapter.h"
#include <boost/bind.hpp>
DisplayAdapter::DisplayAdapter()
	: m_Screen(NULL)
{
	m_ScreenSizes.push_back( new Screen<256,224>() );
	m_ScreenSizes.push_back( new Screen<320, 224>() );
}

void DisplayAdapter::ResetScreen()
{
	std::for_each( m_ScreenSizes.begin(), m_ScreenSizes.end(), boost::bind(&ScreenBase::ResetScreen, _1));
}

std::pair<int,int> DisplayAdapter::GetScreenRes() const
{
	return m_Screen->GetScreenRes();
}

void DisplayAdapter::ApplySTE(int x, int y, STE ste)
{
	m_Screen->ApplySTE(x,y,ste);
}

void* DisplayAdapter::GetScreen() const
{
	return m_Screen->GetScreen();
}

void DisplayAdapter::ResizeScreen( const std::pair<int,int>& size )
{
	for (std::vector<ScreenBase*>::const_iterator it = m_ScreenSizes.begin(),
		itEnd = m_ScreenSizes.end(); it != itEnd; ++it)
	{
		if ((*it)->GetScreenRes() == size)
		{
			m_Screen = *it;
			break;
		}
	}
}

void DisplayAdapter::Begin()
{
	m_Screen->ResetScreen();
	m_Screen->FlipBuffer();
}

void DisplayAdapter::End()
{

}

void DisplayAdapter::WriteToScreen( int x, int y, UBYTE r, UBYTE g, UBYTE b )
{
	m_Screen->WriteToScreen(x,y,r,g,b);
}


DisplayAdapter::ScreenBase::ScreenBase() 
	: m_BufferToggle(true)
{
}

void DisplayAdapter::ScreenBase::FlipBuffer()
{
	m_BufferToggle = !m_BufferToggle;
}
