#pragma once
#ifndef _YM7101_DISPLAY_ADAPTER_H
#define _YM7101_DISPLAY_ADAPTER_H

#include <boost/cstdint.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/function.hpp>

typedef boost::uint8_t UBYTE ;
typedef boost::int8_t SBYTE ;
typedef boost::uint16_t UWORD ;
typedef boost::int16_t SWORD ;
typedef boost::uint32_t UDWORD ;
typedef boost::int32_t SDWORD ;


class DisplayAdapter
{
public:

	enum STE
	{
		STE_NOTSET, STE_SHADOW, STE_NORMAL,  STE_HIGHLIGHT
	};

	DisplayAdapter(void);

	std::pair<int,int> GetScreenRes( ) const ;
	void* GetScreen( ) const;
	void ResizeScreen(const std::pair<int,int>& size);
	void Begin( ) ;
	void End( ) ;
	void WriteToScreen(int x, int y, UBYTE r, UBYTE g, UBYTE b) ;
	void ResetScreen( );
	void ApplySTE( int x, int y, STE ste) ;
private:	
	
	class ScreenBase
	{
	public:
		ScreenBase();
		void FlipBuffer();
		virtual void WriteToScreen(int x, int y, UBYTE r, UBYTE g, UBYTE b) = 0;
		virtual void* GetScreen( ) const = 0;
		virtual void ResetScreen( ) = 0;
		virtual std::pair<int,int> GetScreenRes( ) const = 0;
		virtual void ApplySTE( int x, int y, STE ste) = 0;
	protected:
		bool m_BufferToggle;
	};

	template <int width, int height> 
	class Screen : public ScreenBase
	{
	public:
		virtual void WriteToScreen(int x, int y, UBYTE r, UBYTE g, UBYTE b)
		{
			if (m_BufferToggle)
			{
				m_ScreenBuffer1[y][x][0] = r;
				m_ScreenBuffer1[y][x][1] = g;
				m_ScreenBuffer1[y][x][2] = b;
			}
			else
			{
				m_ScreenBuffer2[y][x][0] = r;
				m_ScreenBuffer2[y][x][1] = g;
				m_ScreenBuffer2[y][x][2] = b;
			}
		}

		virtual void ApplySTE( int x, int y, STE ste)
		{
			if (m_BufferToggle)
			{
				UBYTE r = m_ScreenBuffer1[y][x][0];
				UBYTE g = m_ScreenBuffer1[y][x][1] ;
				UBYTE b = m_ScreenBuffer1[y][x][2] ;
				
				switch(ste)
				{
				case STE_SHADOW:
					r /= 2;
					g /= 2;
					b /= 2;
					break;
				case STE_HIGHLIGHT:
					r = std::min<int>(((255 - r) / 2) + r,255);
					g = std::min<int>(((255 - g) / 2) + g,255);
					b = std::min<int>(((255 - b) / 2) + b,255);
					break;
				}

				m_ScreenBuffer1[y][x][0] = r;
				m_ScreenBuffer1[y][x][1] = g;
				m_ScreenBuffer1[y][x][2] = b;
			}
			else
			{
				UBYTE r = m_ScreenBuffer2[y][x][0];
				UBYTE g = m_ScreenBuffer2[y][x][1] ;
				UBYTE b = m_ScreenBuffer2[y][x][2] ;
				switch(ste)
				{
				case STE_SHADOW:
					r /= 2;
					g /= 2;
					b /= 2;
					break;
				case STE_HIGHLIGHT:
					r = std::min<int>(((255 - r) / 2) + r,255);
					g = std::min<int>(((255 - g) / 2) + g,255);
					b = std::min<int>(((255 - b) / 2) + b,255);
					break;
				}
				m_ScreenBuffer2[y][x][0] = r;
				m_ScreenBuffer2[y][x][1] = g;
				m_ScreenBuffer2[y][x][2] = b;
			}
		}

		virtual void* GetScreen( ) const
		{
			if (m_BufferToggle)
				return (void*)m_ScreenBuffer2;
			else
				return (void*)m_ScreenBuffer1;
		}
		virtual void ResetScreen( )
		{
			if (m_BufferToggle)
				memset(m_ScreenBuffer2, 1,sizeof(m_ScreenBuffer2)) ;
			else
				memset(m_ScreenBuffer1, 1,sizeof(m_ScreenBuffer1)) ;
		}
		virtual std::pair<int,int> GetScreenRes( ) const
		{
			return std::make_pair<int,int>(width,height);
		}
	private:
		UBYTE m_ScreenBuffer1[height][width][3];
		UBYTE m_ScreenBuffer2[height][width][3];
	};
	std::vector<ScreenBase*> m_ScreenSizes;
	ScreenBase* m_Screen;
};

#endif