#pragma once
#ifndef _YM7101_H
#define _YM7101_H

#include <boost/shared_ptr.hpp>
#include "DisplayAdapter.h"

class YM7101
{
public:
	typedef boost::shared_ptr<YM7101> PTR ;
	typedef UWORD (*t_FuncReadMemory16)(UDWORD) ;

	static PTR Create( ) ;

	void Update( int numclicks ) ;

	
	~YM7101(void);
	void WriteControlPort16( UWORD data ) ;
	void WriteDataPort16( UWORD data ) ;
	void WriteControlPort8( UBYTE data ) ;
	void WriteDataPort8( UBYTE data ) ;
	UWORD ReadControlPort16( ) ;
	UWORD ReadDataPort16( ) ;
	UBYTE GetHCounter( ) const ;
	UBYTE GetVCounter( ) const ;
	void SetFuncReadMemory16( const t_FuncReadMemory16& func ) ;
	void DumpVRAM( ) const ;

	bool IsTimeToRender( ) const ;
	UWORD GetHVCounter( ) const ;
	bool IsRequestingInterrupt( ) const ;

	std::pair<int,int> GetScreenResolution( ) const ;

	int GetInterruptType( ) ;
	void* GetScreen( ) const;

private:	
	static const int sc_numregisters = 24 ; 
	static const int sc_numcram = 0x40 ;
	static const int sc_numvsram = 0x40 ;
	
	enum ControlCode
	{
		CC_READ_VRAM = 0x0,
		CC_WRITE_VRAM = 0x1,
		CC_WRITE_REG = 0x2,
		CC_WRITE_CRAM = 0x3,
		CC_READ_VSRAM = 0x4,
		CC_WRITE_VSRAM = 0x5,
		CC_READ_CRAM = 0x8
	} ;
	
	enum PlaneToRender
	{
		PLANE_A,
		PLANE_B,
		PLANE_WINDOW
	};

	YM7101(void);
	void Reset( ) ;
	void UpdateRegister( UWORD data ) ;
	ControlCode GetControlCode() const ;
	void WriteVram( UWORD data ) ;
	void WriteCram( UWORD data ) ;
	void WriteVsram( UWORD data ) ;
	UWORD GetAddressRegister( ) const ;
	UBYTE ReadVram8( ) const ;
	UBYTE ReadCram8( ) const ;
	UBYTE ReadVsram8( ) const ;
	UWORD ReadVram16( ) const ;
	UWORD ReadCram16( ) const ;
	UWORD ReadVsram16( ) const ;
	void Render( ) ;

	UWORD GetSATBase( ) const ;
	UWORD GetPlaneABase( ) const ;
	UWORD GetPlaneBBase( ) const ;
	UWORD GetWindowBase( ) const ;
	bool GetIs40CellMode( ) const ;
	void IncrementAddress( ) ;
	void RenderBackdrop( ) ;
	void RenderSprites( bool highPriority ) ;
	void RenderTiles( const UWORD baseaddress, bool priorityhigh, int hSize, int zSize, UWORD scrollAmount, PlaneToRender plane ) ;
	void RenderWindow( const UWORD baseAddress, bool priorityhigh ) ;
	bool IsRegBitSet(int reg, int bit) const ;
	void DoDMATransfer( UWORD data ) ;	
	void DoDMA68kToVRAM( ) ;
	void DoDMAVRAMFill( UWORD data ) ;
	void DoDMAVRAMCopy( ) ;
	std::pair<int,int> GetPlaneSize() const;
	void ResizeScreen(const std::pair<int,int>& size);
	void GetHScrollAmount(int line, UWORD& aScroll, UWORD& bScroll ) const;
	UWORD GetVScrollAmount( PlaneToRender plane, int doubleColumnNum ) const ;
	UBYTE GetColourShade(UBYTE val) const;
	bool ApplySTE( bool priorityHigh, int xpos, bool isSprite, int spriteColour, int spritePalette );
	bool m_HCounterFirst ;
	bool m_VCounterFirst ;
	UWORD m_VCounter ;
	UWORD m_HCounter ;
	UWORD m_StatusRegister ;
	bool m_Refresh ;

	bool m_RequestInterrupt ;
	UBYTE m_LineInterrupt;

	union Vram
	{
		UBYTE vram8[0x10000] ; // not sure if these numbers are correct
		UWORD vram16[0x8000] ;
	};
	t_FuncReadMemory16 m_FuncReadMemory16;

	UWORD m_CRAM[sc_numcram] ;
	UWORD m_VSRAM[sc_numvsram] ;
	Vram m_VRAM ;
	UBYTE m_Registers[sc_numregisters] ;
	bool m_ControlPortPending ;
	UDWORD m_ControlPortData ;
	DisplayAdapter m_DisplayAdapter;	
	DisplayAdapter::STE m_LineHighlightData[320];	
	int m_MaxSprites ;
	int m_SpritePixelCount ;
	UBYTE m_SpriteLineData[320];
	const int c_maxInt ;
	std::pair<int,std::pair<int,int> > m_SpriteMaskData;
	int m_SpriteMask2 ;
};

#endif