#include "Config_cs_core.h"
#include "IniFileManager.h"
#include "../Utils/Throw.h"
#include <boost/format.hpp>

namespace codeslinger
{
	namespace inifile
	{
		////////////////////////////////////////////////////////////////////////

		IniFileManager::IniFileManager(void)
		{
		}

		////////////////////////////////////////////////////////////////////////

		IniFileManager::~IniFileManager(void)
		{
		}

		////////////////////////////////////////////////////////////////////////

		void IniFileManager::AddIniFile( const std::string& settingName, 
			const boost::shared_ptr<IniFileReader> iniFile )
		{
			VerifyIniFileNotLoaded(settingName) ;
			m_SettingsMap.insert( std::make_pair(settingName, iniFile) ) ;
		}

		////////////////////////////////////////////////////////////////////////

		bool IniFileManager::IsIniFileLoaded( const std::string& iniFile ) const
		{
			return m_SettingsMap.find(iniFile) != m_SettingsMap.end() ;
		}

		////////////////////////////////////////////////////////////////////////

		void IniFileManager::VerifyIniFileLoaded( const std::string& iniFile ) const
		{			
			if (!IsIniFileLoaded(iniFile))
			{
				THROW( ( boost::format("Ini file %1% Not Loaded") % iniFile ).str() ) ;
			}
		}

		////////////////////////////////////////////////////////////////////////

		void IniFileManager::VerifyIniFileNotLoaded( const std::string& iniFile ) const
		{
			if (IsIniFileLoaded(iniFile))
			{
				THROW( ( boost::format("Ini file %1% Already Loaded") % iniFile ).str() ) ;
			}
		}

		////////////////////////////////////////////////////////////////////////

		boost::shared_ptr<IniFileReader> IniFileManager::GetIniFileReader( const std::string& iniFile ) const
		{
			boost::shared_ptr<IniFileReader> res ;
			if (IsIniFileLoaded(iniFile))
			{
				res = (*m_SettingsMap.find(iniFile)).second ;
			}
			return res ;
		}

		////////////////////////////////////////////////////////////////////////

	};
};