#pragma once
#ifndef _CS_INI_FILE_MANAGER_H
#define _CS_INI_FILE_MANAGER_H

#include <cs_core/Utils/Singleton.h>
#include <cs_core/IniFiles/IniFileReader.h>
#include <string>
#include <map>
#include <boost/shared_ptr.hpp>

namespace codeslinger
{
	namespace inifile
	{
		class IniFileManager : public codeslinger::Singleton<IniFileManager>
		{
		public:
			void AddIniFile( const std::string& settingName, 
				const boost::shared_ptr<IniFileReader> iniFile ) ;

			template <typename T>
			bool GetValue( const std::string& iniFile,
				const std::string& settingName, T& out) const
			{
				VerifyIniFileLoaded(iniFile) ;
				boost::shared_ptr<IniFileReader> iniReader = GetIniFileReader(iniFile) ;
				return iniReader->GetValue<T>(settingName, out) ;
			}

		private:

			friend class codeslinger::Singleton<IniFileManager> ;

			virtual ~IniFileManager ( ) ;
			IniFileManager	( ) ;	

			bool IsIniFileLoaded( const std::string& iniFile ) const ;
			void VerifyIniFileLoaded( const std::string& iniFile ) const ;
			void VerifyIniFileNotLoaded( const std::string& iniFile ) const ;
			boost::shared_ptr<IniFileReader> GetIniFileReader(const std::string& iniFile) const ;

			typedef std::map< std::string, boost::shared_ptr<IniFileReader> > SETTINGSMAP ;
			SETTINGSMAP m_SettingsMap ;
		};
	};
};

#endif