#include "Config_cs_core.h"
#include "IniFileReader.h"
#include "IniFileManager.h"
#include <fstream>
#include <algorithm>
#include <boost/format.hpp>

namespace codeslinger
{
	namespace inifile
	{
		///////////////////////////////////////////////////////////////////////

		IniFileReader::IniFileReader(const std::string& folderName, const std::string& filename)
		{
			ReadIniFile( (boost::format("%1%/%2%") % folderName % filename).str() ) ;
		}

		///////////////////////////////////////////////////////////////////////

		IniFileReader::~IniFileReader(void)
		{
		}

		///////////////////////////////////////////////////////////////////////

		void IniFileReader::FlushSettings( )
		{
			m_Settings.clear( ) ;
		}

		///////////////////////////////////////////////////////////////////////

		bool IniFileReader::GetSettingValue(const std::string& settingName, std::string& value) const
		{
			codeslinger::string::ci_string ciSettingName = settingName.c_str() ;
			SETTINGS::const_iterator it = m_Settings.find(ciSettingName) ;

			if (it != m_Settings.end())
			{
				value = (*it).second ;
			}
			
			return it != m_Settings.end();
		}

		///////////////////////////////////////////////////////////////////////

		void IniFileReader::ReadIniFile( const std::string& filename )
		{			
			typedef std::string::size_type size_type ;

			std::ifstream file(filename.c_str()) ;

			if (file.is_open())
			{
				while(!file.eof())
				{
					std::string line ;
					std::getline( file, line) ;
					
					const size_type seperatorpos = line.find('=') ;
					if (std::string::npos != seperatorpos)
					{
						size_type endpos = line.size() ;
						size_type commentpos = line.find('*') ;

						if (std::string::npos != commentpos)
						{
							endpos = commentpos ;
						}

						std::string settingName(line, 0,seperatorpos) ;
						std::string settingValue(line, seperatorpos+1, endpos-(seperatorpos+1)) ;

						settingName.erase( std::remove(settingName.begin(), settingName.end(), ' '), settingName.end()) ;
						settingValue.erase( std::remove(settingValue.begin(), settingValue.end(), ' '), settingValue.end()) ;

						if (settingName == "#IniFile")
						{
							ReadIniFile(settingValue) ;
						}
						else
						{
							m_Settings.insert(std::make_pair(settingName.c_str(), settingValue)) ;
						}
					}
				}

				file.close() ;
			}
		}

		///////////////////////////////////////////////////////////////////////
	};

};