#pragma once
#ifndef _CS_INI_FILE_READER_H
#define _CS_INI_FILE_READER_H

#include <cs_core/Strings/ci_string.h>
#include <string>
#include <map>
#include <sstream>
#include <boost/enable_shared_from_this.hpp>

namespace codeslinger
{
	namespace inifile
	{
		class IniFileReader
		{
		public:
			IniFileReader(const std::string& folderName, const std::string& filename) ;
			~IniFileReader(void); 

			void FlushSettings( ) ;
			void ReadIniFile( const std::string& filename ) ;

			template <typename T>
			bool GetValue( const std::string& settingName, T& res) const
			{
				std::string value ;
				bool found = GetSettingValue(settingName, value) ;
				if (found)
				{
					std::stringstream ss ;
					ss << value ;
					ss >> res ;
				}
				else
				{
					res = T( ) ;
				}
				return found ;
			}

			template <>
			bool GetValue( const std::string& settingName, std::string& res) const
			{
				return GetSettingValue( settingName, res ) ;
			}

		private:
			typedef std::map<codeslinger::string::ci_string, std::string> SETTINGS ;
			SETTINGS m_Settings ;

			bool GetSettingValue( const std::string& settingName, std::string& value) const ;
		};
	};
};

#endif