#pragma once
#ifndef _CS_INI_FILES_H
#define _CS_INI_FILES_H

#include "IniFileReader.h"
#include "IniFileManager.h"

namespace codeslinger
{
	namespace inifile
	{
		boost::shared_ptr<IniFileReader> IniFileLoadSettings( 
			const std::string& settingName, const std::string& filename) ;

		/////////////////////////////////////////////////////////////////////////

		IniFileManager* IniFileCreateManager( ) ;

		/////////////////////////////////////////////////////////////////////////

		void IniFileDestroyManager( ) ; 

		/////////////////////////////////////////////////////////////////////////

		template <typename T>
		void IniFileGetValue( const std::string& iniFile, const std::string& settingName, T& res )
		{
			bool found = IniFileManager::GetSingletonPtr()->GetValue<T>(iniFile, settingName, res) ;
			if (!found)
			{
				THROW( ( boost::format("No setting found %1% in Ini file %2%") % settingName% iniFile ).str() ) ;
			}
		}

		/////////////////////////////////////////////////////////////////////////

		template <typename T>
		void IniFileGetValue( const std::string& settingName, T& res )
		{
			return IniFileGetValue<T>("Main", settingName, res) ;
		}

		/////////////////////////////////////////////////////////////////////////

		template <typename T>
		void IniFileGetValueDefault( const std::string& iniFile, const std::string& settingName, T& res,  const T& def )
		{
			bool found = IniFileManager::GetSingletonPtr()->GetValue<T>(iniFile, settingName, res) ;
			if (!found)
			{
				res = def ;
			}
		}

		/////////////////////////////////////////////////////////////////////////

		template <typename T>
		void IniFileGetValueDefault( const std::string& settingName, T& res, const T& def )
		{
			return IniFileGetValueDefault<T>("Main", settingName, res, def) ;
		}

		/////////////////////////////////////////////////////////////////////////

		template <typename T>
		T IniFileReturnDefault( const std::string& settingName, const T& def )
		{
			T res ;
			IniFileGetValueDefault<T>("Main", settingName, res, def) ;
			return res ;
		}

		/////////////////////////////////////////////////////////////////////////

		template <typename T>
		T IniFileReturnDefault( const std::string& iniFilename, const std::string& settingName, const T& def )
		{
			T res ;
			IniFileGetValueDefault<T>(iniFilename, settingName, res, def) ;
			return res ;
		}

		/////////////////////////////////////////////////////////////////////////
	};
};

#endif