#include "Config_cs_core.h"
#include "Logfile.h"
#include "LoggingManager.h"
#include "../IniFiles/IniFiles.h"
#include <algorithm>
#include <functional>
#include <boost/bind.hpp>
#include <boost/format.hpp>
#include <boost/filesystem.hpp>
namespace codeslinger
{
	namespace logging
	{
		////////////////////////////////////////////////////////////////////

		Logfile::Logfile(void)
		{
		}

		////////////////////////////////////////////////////////////////////

		Logfile::Logfile(const std::string& filename)
			: m_Name(filename)
		{
			codeslinger::inifile::IniFileGetValueDefault<bool>( (boost::format("LogFile%1%Enabled")
				% filename ).str(), m_Enabled, true ) ;
			codeslinger::inifile::IniFileGetValueDefault<int>( (boost::format("LogFile%1%Level") % 
				filename ).str(), m_Level, 0 ) ;
			if (m_Enabled)
			{
				boost::filesystem::create_directory("Logs");
				m_File.open( (boost::format("Logs/%1%.log") % filename ).str().c_str() ) ;
			}
		}

		////////////////////////////////////////////////////////////////////

		Logfile::~Logfile(void)
		{
			if (m_File.is_open())
			{
				m_File.close() ;
			}
		}

		////////////////////////////////////////////////////////////////////
		
		void Logfile::AddDestination(const boost::shared_ptr<LogMsgDestination>& destination)
		{
			m_Destinations.push_back(destination) ;
		}

		////////////////////////////////////////////////////////////////////
		
		void Logfile::LogMessage(const std::string& message)
		{
			ProcessMessage(message) ;
		}

		////////////////////////////////////////////////////////////////////

		void Logfile::ProcessMessage(const std::string& message)
		{
			if (m_File.is_open())
			{
				m_File << message.c_str() << std::endl ;
			}
			
			std::for_each( m_Destinations.begin(), m_Destinations.end(), 
				boost::bind(&LogMsgDestination::ProcessMessage, _1, message) ) ;
		}

		////////////////////////////////////////////////////////////////////

		bool Logfile::IsEnabled() const
		{
			return m_Enabled ;
		}

		////////////////////////////////////////////////////////////////////

		int Logfile::GetLevel() const
		{
			return m_Level ;
		}

		////////////////////////////////////////////////////////////////////

		std::string Logfile::GetName( ) const
		{
			return m_Name ;
		}

		////////////////////////////////////////////////////////////////////
	};
};