#include "Config_cs_core.h"
#include "Logging.h"
#include "LoggingManager.h"
#include "Logfile.h"
#include <stdarg.h>

namespace codeslinger
{
	namespace logging
	{
		////////////////////////////////////////////////////////////////

		void LoggingCreateManager( )
		{
			LoggingManager::CreateSingleton( ) ;
		}

		////////////////////////////////////////////////////////////////

		void LoggingDestroyManager( )
		{
			LoggingManager::DestroySingleton( ) ;
		}

		////////////////////////////////////////////////////////////////

		void LogMessage(const std::string& subset, const std::string& msg)
		{
			LogMessage("Main", subset, 0, msg) ;
		}

		////////////////////////////////////////////////////////////////

		void LogMessage(const std::string& subset, int level, const std::string& msg)
		{
			LogMessage("Main", subset, level, msg) ;
		}
		
		////////////////////////////////////////////////////////////////

		void LogMessage(const std::string& groupName, 
			const std::string& subset, const std::string& msg)
		{
			LogMessage(groupName, subset, 0, msg) ;
		}

		////////////////////////////////////////////////////////////////

		void LogMessage(const std::string& groupName, 
			const std::string& subset, int level, const std::string& msg)
		{
			LoggingManager::GetSingletonPtr()->LogMessage(groupName, subset, level, msg) ;
		}

		////////////////////////////////////////////////////////////////

		void LoggingAddNewGroup( const std::string& groupName )
		{
			LoggingManager::GetSingletonPtr()->AddNewGroup(groupName);
			LoggingChainLogFiles(groupName, "Main");
		}

		////////////////////////////////////////////////////////////////

		void LoggingAddNewSubset( const std::string& subset )
		{
			LoggingManager::GetSingletonPtr()->AddNewSubset(subset);
		}
		
		////////////////////////////////////////////////////////////////

		void LoggingAddDestination( const std::string& groupName, const std::string& subset,
			const boost::shared_ptr<LogMsgDestination>& dest )
		{
			LoggingManager::GetSingletonPtr()->GetLogfile(groupName, subset)->AddDestination(dest);
		}

		////////////////////////////////////////////////////////////////

		void LoggingChainLogFiles( const std::string& source, const std::string& destination )
		{
			LoggingManager::GetSingletonPtr()->MergeGroupSubsets(source, destination);
		}

		////////////////////////////////////////////////////////////////

		bool LoggingDoesLogFileExist(const std::string& groupName, const std::string& subset)
		{
			bool res = true ;
			try
			{
				LoggingManager::GetSingletonPtr()->GetLogfile(groupName, subset) ;
			}
			catch (const std::exception&)
			{
				res = false ;
			}
			return res ;
		}

		////////////////////////////////////////////////////////////////

	};
};