#pragma once
#ifndef _CS_LOGGING_MANAGER_H
#define _CS_LOGGING_MANAGER_H

#include <cs_core/Utils/Singleton.h>
#include <string>
#include <map>
#include <set>
#include <boost/shared_ptr.hpp>

namespace codeslinger
{
	namespace logging
	{
		class Logfile ;
		class LogMsgDestination ;

		class LoggingManager : public codeslinger::Singleton<LoggingManager>
		{
		private:
			typedef std::map< std::string, boost::shared_ptr<Logfile> > LOG_FILE_MAP ;
		public:
			void AddLogFile( const std::string& groupName, const std::string& subsetName, 
				const boost::shared_ptr<Logfile>& logfile ) ;
			void LogMessage( const std::string& groupName, const std::string& subsetName, 
				int level, const std::string& message ) ;
			void AddNewGroup(const std::string& name);
			boost::shared_ptr<Logfile> GetLogfile( const std::string& groupName, 
				const std::string& subsetName ) const ;
			void MergeGroupSubsets( const std::string& source, const std::string& destination ) ;
			void AddNewSubset( const std::string& subset ) ;
			void SetEnabled( bool enabled );
			bool GetEnabled( ) const ;

		private:
			friend class codeslinger::Singleton<LoggingManager> ;

			virtual ~LoggingManager ( ) ;
			LoggingManager	( ) ;	

			std::string GetTimeStamp( ) const ;

			LOG_FILE_MAP m_LogfileMap ;

			std::set<std::string> m_sublayers;
			std::set<std::string> m_groups;
			bool m_enabled;
		};
	};
};


#endif
