/////////////////////////////////////////////////////////////////

// if we are in debug then trigger a hardware breakpoint if the condition is not true.
// because it is in a macro the application will break on the line that called ASSERT and not
// deep in the callstack like assert() does.

// by defining a no-op assert macro for release builds it means i can use the ASSERT macro
// like this:
// if (someCondition)
//     ASSERT(false) ;

// rather than using it like this
// if (someConditioN)
// {
//  ASSERT(false) ;
// }

#ifndef _CODESLINGER_DEBUGGING_H
#define _CODESLINGER_DEBUGGING_H

#ifdef WIN32

#ifdef _DEBUG
#define ASSERT( condition )  if (!condition) _asm int 3 ;
#else
#define ASSERT(condition) _asm nop ;
#endif

#else
#include <assert.h>
#define ASSERT( condition ) assert(condition) ;

#endif



///////////////////////////////////////////////////////////////////////

// the following will use the preprocessor to generate a compile time string
// which holds the current file, function and line number

#define STRINGIFY(x) #x
#define TOSTRING(x) STRINGIFY(x)

#ifdef WIN32
#define SOURCEINFO  "["  __FILE__  " : "  __FUNCTION__  " : " TOSTRING(__LINE__) "] "
#else
#define SOURCEINFO  "["  __FILE__  " : " TOSTRING(__LINE__) "] "
#endif

#endif