#ifndef _MATHS_H
#define _MATHS_H

namespace codeslinger
{
	namespace maths
	{
		//////////////////////////////////////////////////////////////////////////////////

		template< typename typeData >
		inline bool TestBit( typeData inData, int inBitPosition )
		{
			typeData lMsk = 1 << inBitPosition ;
			return ( inData & lMsk ) ? true : false ;
		}

		//////////////////////////////////////////////////////////////////////////////////

		template< typename typeData >
		inline typeData BitGet( typeData inData, int inBitPosition )
		{
			typeData lMsk = 1 << inBitPosition ;
			return lMsk ;
		}

		//////////////////////////////////////////////////////////////////////////////////

		template< typename typeData >
		inline typeData BitGetVal( typeData inData, int inBitPosition )
		{
			typeData lMsk = 1 << inBitPosition ;
			return ( inData & lMsk ) ? 1 : 0 ;
		}

		//////////////////////////////////////////////////////////////////////////////////

		template< typename typeData >
		inline typeData BitSet( typeData inData, int inBitPosition )
		{
			typeData lMsk = 1 << inBitPosition ;
			inData |= lMsk ;
			return inData ;
		}

		//////////////////////////////////////////////////////////////////////////////////

		template< typename typeData >
		inline typeData BitReset( typeData inData, int inBitPosition )
		{
			typeData lMsk = 1 << inBitPosition ;
			inData &= ~lMsk ;
			return inData ;
		}

		//////////////////////////////////////////////////////////////////////////////////

		template < typename typedata >
		inline typedata BitSetAll( typedata inNumBits )
 		{
/* 			if (inNumBits > 32)
 				throw std::runtime_error("Num bits cannot exceed 32 for inNumBits");
 			else */if (inNumBits == 32)
				return 0xFFFFFFFF;
			typedata test = typedata(typedata(1) << inNumBits);
			return test - 1 ;
		}

		//////////////////////////////////////////////////////////////////////////////////

		template < typename typedata >
		inline typedata BitRotateLeft( typedata inData, int inNumToRotate, int inNumBits )
		{
			typedata res = ( inData << inNumToRotate ) & BitSetAll<typedata>( inNumBits ) ;
			typedata lower = inData >> (inNumBits - inNumToRotate) & BitSetAll<typedata>( inNumToRotate ) ;
			res &= BitSetAll<typedata>( inNumBits - inNumToRotate) << inNumToRotate ;
			res |= lower ;
			return res ;
			//return (inData << inNumToRotate) | (inData >> (inNumBits - inNumToRotate));
		}

		//////////////////////////////////////////////////////////////////////////////////

		template < typename typedata >
		inline typedata BitRotateRight( typedata inData, int inNumToRotate, int inNumBits )
		{
			typedata res = ( inData >> inNumToRotate ) & BitSetAll<typedata>( inNumBits - inNumToRotate ) ;
			typedata upper = ( inData & BitSetAll<typedata>( inNumToRotate ) ) << (inNumBits-inNumToRotate) ;
			res |= upper ;
			return res ;
			//return (inData >> inNumToRotate) | (inData << (inNumBits - inNumToRotate));

		}

		//////////////////////////////////////////////////////////////////////////////////

		template< typename typeData >
		inline typeData BitConditionalSet( bool condition, typeData inData, int inBitPosition )
		{
			if (condition)
				return BitSet<typeData>( inData, inBitPosition ) ;
			return BitReset<typeData>( inData, inBitPosition ) ;
		}

		//////////////////////////////////////////////////////////////////////////////////

		template< typename typeData >
		int BitCount (typeData inData, int totalBits )
		{
			int res = 0 ;
			for (int i = 0 ; i < totalBits; i++)
			{
				if (TestBit<typeData>(inData,i))
					res++ ;
			}
			return res ;
		}

		typedef boost::uint8_t UBYTE ;
		typedef boost::int8_t SBYTE ;
		typedef boost::uint16_t UWORD ;
		typedef boost::int16_t SWORD ;
		typedef boost::uint32_t UDWORD ;
		typedef boost::int32_t SDWORD ;


		//////////////////////////////////////////////////////////////////////////////////

		void SwapEndian( boost::uint16_t& toSwap ) ;
		

		//////////////////////////////////////////////////////////////////////////////////



		void SwapEndian( boost::uint32_t& toSwap ) ;
		

		//////////////////////////////////////////////////////////////////////////////////

		template <typename T>
		inline T modulus_neg(T a, T b)
		{
			T const rem = a % b;
			if (-1 % 2 == 1)
				return rem;
			else
				return rem < 0? rem + b : rem;

		}

		//////////////////////////////////////////////////////////////////////////////////
	};
};

#endif