#pragma once
#ifndef _CARTRIDGE_H
#define _CARTRIDGE_H

#include <string>
#include <boost/shared_ptr.hpp>
#include <boost/shared_array.hpp>
#include <boost/filesystem.hpp>
#include <boost/logic/tribool.hpp>

class MegaDriveCartridge
{
	struct RomHeader
	{         
		UBYTE segatitle[0x10]; //0x100 
		UBYTE copyright[0x10]; //0x110
		UBYTE domestic[0x30];  //0x120 
		UBYTE overseas[0x30];  //0x150 
		UBYTE production[0xe]; //0x180
		UWORD checksum;       //0x18e
		UBYTE control[0x10];    //0x190
		UDWORD rom_start; //0x1a0
		UDWORD rom_end; //0x1a4
		UDWORD ram_start; //0x1a8
		UDWORD ram_end; //0x1ac
		UBYTE sram[0xc];//0x1b0
		UBYTE modem[12];    //0x1bc  
		UBYTE memo[40];       
		UBYTE country[16];    
	} ;

public:
	typedef boost::shared_ptr<MegaDriveCartridge> PTR ;

	static PTR Create( ) ;

	~MegaDriveCartridge(void);

	bool Init( const boost::filesystem::path& romFile ) ;
	bool ValidateChecksum( ) const ;
	void FixChecksum( ) ;
	void PageMemory( int pagenum, unsigned long pagesize, 
		unsigned long destStartPos, boost::shared_array<UBYTE>& destination ) ;
	void WriteSaveRam( UDWORD address, UBYTE data ) ;
	bool IsSaveRamAvailable( ) const ;
	std::pair<UDWORD,UDWORD> GetSaveRamRange() const;
	UBYTE ReadSaveRam( UDWORD address ) const;
	void BackupSaveRam( ) ;

private:
	MegaDriveCartridge(void) ;

	bool LoadROMFile( const boost::filesystem::path& romFile ) ;
	bool LoadBinFormat( const boost::filesystem::path& romFile ) ;
	bool LoadMDFormat( const boost::filesystem::path& romFile ) ;
	bool LoadSMDFormat( const boost::filesystem::path& romFile ) ;
	bool LoadSplitSMDFormat( const boost::filesystem::path& romFile ) ;
	bool LoadZipFormat( const boost::filesystem::path& romFile ) ;

	boost::tribool LoadSMDFile( const boost::filesystem::path& romFile, 
		boost::shared_array<UBYTE>& destination, long& memorySize ) ;
	void LoadROMHeader( ) ;
	void DumpROMHeader( ) ;
	bool ValidateROMHeader( ) const ;
	UWORD CalculateChecksum( ) const ;
	void CreateSaveRam( );
	void LoadSaveRam( ) ;

	boost::shared_array<UBYTE> m_CartridgeMemory ;
	boost::shared_array<UBYTE> m_SaveRam ;
	long m_CartridgeSize ;
	RomHeader m_RomHeader ;
	bool m_SaveRamAvailable;
	UDWORD m_SaveRamStart;
	UDWORD m_SaveRamEnd;
	UDWORD m_SaveRamLength;
};

#endif