#include "Config_MegaDrive.h"
#include "MDDebugger.h"
#ifdef ENABLE_DEBUGGER
#include "InputDevice/iInputDevice.h"
#include "MegaDrive.h"
#include "boost/bind.hpp"
#include <YM7101/YM7101.h>
#include <cs_core/Logging/LoggingManager.h>

/////////////////////////////////////////////////////////////////////////////////

MDDebugger::MDDebugger(const boost::shared_ptr<MegaDrive>& megaDrive, const boost::shared_ptr<iInputDevice>& input)
	: m_MegaDrive(megaDrive)
	, m_InputDrive(input)
{
	m_InputDrive->SetFunctionKeyPress(boost::bind(
		&MDDebugger::HandleKeyEvent, this, _1, _2 ) ) ;
}

/////////////////////////////////////////////////////////////////////////////////

void MDDebugger::Update()
{
	m_InputDrive->Update();
}

/////////////////////////////////////////////////////////////////////////////////

void MDDebugger::HandleKeyEvent(bool down, int id)
{
	if (down && id < DEBUG_NUM)
	{
		switch(id)
		{
		case DEBUG_PAUSE:
			m_MegaDrive->SetPause( true ) ;//!m_MegaDrive->IsPaused() );
			break;
		case DEBUG_DUMP_VRAM: 
			m_MegaDrive->m_YM7101->DumpVRAM();
			break;
		case DEBUG_TOGGLE_LOGGING_ENABLED:
			{
				codeslinger::logging::LoggingManager* lman = codeslinger::logging::LoggingManager::GetSingletonPtr();
				lman->SetEnabled( !lman->GetEnabled() ) ;
			}			
			break;
		default:
			break;
		}
		
	}
}

/////////////////////////////////////////////////////////////////////////////////


#endif