#include "Config_MegaDrive.h"
#include "SDLGLWindow.h"
#include <cs_core/IniFiles/IniFiles.h>
#include <SDL/SDL.h>
#include <SDL/SDL_opengl.h>
#include <GL/glut.h>
#include <stdlib.h>
//////////////////////////////////////////////////

iOutputDevice::PTR SDLGLWindow::Create()
{
	SDLGLWindow* rawPtr = new SDLGLWindow() ;
	rawPtr->CreateSDLWindow( ) ;
	iOutputDevice::PTR res = iOutputDevice::PTR( rawPtr ) ;
	return res ;
}

//////////////////////////////////////////////////

SDLGLWindow::SDLGLWindow(void)
	: m_Width(0)
	, m_Height(0)
	, m_TextureCreated(false)
{	
}

//////////////////////////////////////////////////

SDLGLWindow::~SDLGLWindow(void)
{
}

//////////////////////////////////////////////////

void SDLGLWindow::SetVideoBufferSize(int width, int height)
{
	if (m_Width == width && m_Height == height )
	{
		return ;
	}

	m_Width = width ;
	m_Height = height ;
	
	Uint32 flags = SDL_OPENGL | SDL_RESIZABLE ;

	bool fullscreen = codeslinger::inifile::IniFileReturnDefault<bool>("Fullscreen", false) ;
	int fullscreenWidth = codeslinger::inifile::IniFileReturnDefault<int>("FullscreenWidth", 800) ;
	int fullscreenHeight = codeslinger::inifile::IniFileReturnDefault<int>("FullscreenHeight",600) ;

	int scale = 1 ; // just change me to scale
	int windowwidth = m_Width * scale;
	int windowheight = m_Height * scale;

	if (fullscreen)
	{
		windowwidth = fullscreenWidth ;
		windowheight = fullscreenHeight ;
		flags |= SDL_FULLSCREEN ;
	}

	if( SDL_SetVideoMode( windowwidth, windowheight, 16, flags ) == NULL )
	{
		LogMessage("err", "ERROR: Failed to SDL_SetVideoMode") ;
		return ;
	}

	InitGL( ) ;
}

//////////////////////////////////////////////////

void SDLGLWindow::CreateSDLWindow( )
{
	if( SDL_Init( SDL_INIT_EVERYTHING ) < 0 )
	{
		LogMessage("err", "ERROR: Failed to initialize SDL window") ;
		return ;
	}

	SDL_WM_SetCaption( "Sega Mega Drive", NULL );
	SDL_WM_SetIcon(SDL_LoadBMP("joypad.bmp"), NULL);
}

//////////////////////////////////////////////////

void SDLGLWindow::InitGL( )
{
	if (m_TextureCreated)
	{
		glDeleteTextures(1, &m_TextureId) ;
	}

	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	glGenTextures(1, &m_TextureId);
	m_TextureCreated = true ;

	glBindTexture(GL_TEXTURE_RECTANGLE_ARB,m_TextureId); 
	glEnable( GL_TEXTURE_RECTANGLE_ARB );

	glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameterf(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGB,
		m_Width, m_Height, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);

	glMatrixMode(GL_PROJECTION); 
	glLoadIdentity ();
	gluOrtho2D(0.0, 1.0, 1.0, 0.0);
}

//////////////////////////////////////////////////

void SDLGLWindow::Render( void* videobuffer ) const
{
	glLoadIdentity ();
	gluOrtho2D(0.0, 1.0, 1.0, 0.0);

	glClear(GL_COLOR_BUFFER_BIT);

	glTexSubImage2D(GL_TEXTURE_RECTANGLE_ARB, 0,
		0,0,m_Width,m_Height,
		GL_RGB, GL_UNSIGNED_BYTE, (const GLvoid *)videobuffer);

	glColor4f(1.0, 1.0, 1.0, 1.0);
	glBegin(GL_QUADS);
		glTexCoord2f(0, 0); glVertex2f(0, 0);
		glTexCoord2f(m_Width,0); glVertex2f(1, 0);
		glTexCoord2f(m_Width,m_Height); glVertex2f(1, 1);
		glTexCoord2f(0, m_Height); glVertex2f(0, 1);
	glEnd();

	SDL_GL_SwapBuffers();
}

//////////////////////////////////////////////////

bool SDLGLWindow::Update( )
{
 	SDL_Event event;
 	SDL_PollEvent( &event ) ;
 	
 	if (event.type == SDL_VIDEORESIZE)
 	{
 		SDL_SetVideoMode( event.resize.w, event.resize.h, 16, SDL_OPENGL | SDL_RESIZABLE ) ;
 		InitGL( ) ;
 	}
	else if (event.type == SDL_QUIT)
	{
		return false;
	}
	return true;
}
